(***** Exercice 3 *****)

(*** Question 1 ***)
type formule  =
| Vrai
| Faux
| Var of string
| Non of formule
| Et of formule * formule
| Ou of formule * formule
| Impl of formule * formule

(*** Question 2 ***)
let a = Var "a"
let b = Var "b"
let c = Var "c"

let f1 = Ou (a, Et (b, c))
let f2 = Ou (Et (a, Non b), Et (b, Non (Ou (c, a))))
let f3 = Ou (Non a, Impl (a, b)) 
let f4 = Et (a, Et (Impl (b, c), Impl (c, b)))



(***** Exercice 4 *****)

type valuation = (string * bool) list


(*** Question 1 ***)
let rec evalue_var (v: valuation) (x: string) : bool =
  match v with 
  | [] -> failwith "Variable non définie dans la valuation"
  | (y, b) :: v' when y = x -> b
  | _ :: v' -> evalue_var v' x

let rec evalue (v: valuation) (f: formule) : bool =
  match f with
  | Vrai -> true
  | Faux -> false
  | Var x -> evalue_var v x
  | Non f' -> not (evalue v f')
  | Et (f1, f2) -> evalue v f1 && evalue v f2
  | Ou (f1, f2) -> evalue v f1 || evalue v f2
  | Impl (f1, f2) -> if evalue v f1 then evalue v f2 else true


(*** Question 2 ***)
let variables (f: formule) : string list = 
  (* Fonction auxiliaire qui effectue un parcours en profondeur sur la formule.
     l accumule les nom variables rencontrés jusqu'à présent.
     La fonction renvoie les variables de l auxquels on ajoute les nouvelles variables rencontrées. *)
  let rec aux f l =
    match f with 
    | Vrai -> l
    | Faux -> l
    | Var x -> if List.mem x l then l else x :: l
    | Non f' -> aux f' l
    | Et (f1, f2) -> aux f2 (aux f1 l)
    | Ou (f1, f2) -> aux f2 (aux f1 l)
    | Impl (f1, f2) -> aux f2 (aux f1 l)
  in List.rev (aux f []) (* List.rev facultatif, permet d'obtenir les variables dans l'ordre naturel de lecture. *)


(* quelques tests *)
let () = 
  assert(variables f1 = ["a"; "b"; "c"]);
  assert(variables f2 = ["a"; "b"; "c"]);
  assert(variables f3 = ["a"; "b"]);
  assert(variables f4 = ["a"; "b"; "c"])



(*** Question 3 ***)
exception DerniereValuation

let rec next (v: valuation) : valuation =
  match v with
  | [] -> raise DerniereValuation
  | (x, false) :: v' -> (x, true) :: v'
  | (x, true) :: v' -> (x, false) :: next v'

(* quelques tests *)
let _ = 
  let v000 = [("a", false); ("b", false); ("c", false)] in
  let v001 = next v000 in
  let v010 = next v001 in
  let v111 = next (next (next (next (next v010)))) in
  assert(v001 = [("a", true);  ("b", false); ("c", false)]);
  assert(v010 = [("a", false); ("b", true);  ("c", false)]);
  assert(v111 = [("a", true);  ("b", true);  ("c", true)])



(*** Question 4 ***)
(* Cette fonction renvoie Some v si v satisfait la formule, None si la formule n'est pas satisfiable. *)
let bruteforce (f: formule) : valuation option =
  let rec loop v =
    if evalue v f then Some v else loop (next v)
  in
  try 
    let var = variables f in 
    (* Transforme la liste var en la première valuation. *)
    let v = List.map (fun x -> (x, false)) var in
    loop v
  with
  | DerniereValuation -> None


let v1 = bruteforce f1
let v2 = bruteforce f2
let v3 = bruteforce f3
let v4 = bruteforce f4


(*** Question 5 ***)
let bruteforce_nb (f: formule) : int =
  let compteur = ref 0 in 
  let rec loop v =
    if evalue v f then incr compteur; 
    loop (next v)
  in
  try 
    let var = variables f in 
    (* Transforme la liste var en la première valuation. *)
    let v = List.map (fun x -> (x, false)) var in
    loop v
  with
  | DerniereValuation -> !compteur

let nb1 = bruteforce_nb f1
let nb2 = bruteforce_nb f2
let nb3 = bruteforce_nb f3
let nb4 = bruteforce_nb f4


(*** Question 6 ***)
let tautologie (f: formule) : bool =
  let rec loop v =
    if evalue v f then loop (next v) else false
  in
  try 
    let var = variables f in 
    (* Transforme la liste var en la première valuation. *)
    let v = List.map (fun x -> (x, false)) var in
    loop v
  with
  | DerniereValuation -> true (* Si on atteint la dernière, c'est une tautologie *)

  
let est_tautologie_1 = tautologie f1
let est_tautologie_2 = tautologie f2
let est_tautologie_3 = tautologie f3
let est_tautologie_4 = tautologie f4
let est_tautologie_5 = tautologie (Ou (a, Non a))
let est_tautologie_6 = tautologie Vrai
let est_tautologie_7 = tautologie Faux



(*** Question 7 ***)
let rec pretty_print (v: valuation) =
  match v with
  | [] -> print_newline ()
  | (x, b) :: v' -> 
      print_string x; 
      print_string " --> "; 
      print_int (if b then 1 else 0); 
      print_string "  ";
      pretty_print v'


(*** Question 8 ***)

(* La fonction renvoie None si la formule est une tautologie, et Some v si v ne satisfait pas la formule *)
let tautologie_contre_exemple (f: formule) : valuation option =
  bruteforce (Non f)

let contre_exemple_1 = tautologie_contre_exemple f1
let contre_exemple_2 = tautologie_contre_exemple f2
let contre_exemple_3 = tautologie_contre_exemple f3
let contre_exemple_4 = tautologie_contre_exemple f4
let contre_exemple_5 = tautologie_contre_exemple (Ou (a, Non a))
let contre_exemple_6 = tautologie_contre_exemple Vrai
let contre_exemple_7 = tautologie_contre_exemple Faux


